/*
 * ReversiAI.java
 *
 * Last modified August 8, 2005
 *
 * Artificial Intelligence implementing minimax with alpha-beta
 * cutoffs for othello/reversi game.
 */

import java.awt.*;
import java.util.*;

/**
 *
 * @author Jaime Prez, Toms Aguado, ngel Jara
 */
public class ReversiAI {
    
    private static byte[][] weights = {
        {120, -20, 20,  5,  5, 20, -20, 120},
        {-20, -40, -5, -5, -5, -5, -40, -20},
        { 20,  -5, 15,  3,  3, 15,  -5,  20},
        {  5,  -5,  3,  3,  3,  3,  -5,   5},
        {  5,  -5,  3,  3,  3,  3,  -5,   5},
        { 20,  -5, 15,  3,  3, 15,  -5,  20},
        {-20, -40, -5, -5, -5, -5, -40, -20},
        {120, -20, 20,  5,  5, 20, -20, 120}
    };

    private byte plys;
    private byte uplys;
    private ReversiTree tree;
    private byte turn;
    private Point humanMove;
    private boolean singlePlayer = true;
    private ReversiGUI gui;
    
  
    /** Creates a new instance of ReversiAI
     *  By default, a 8x8 board and the most simple
     *  depth for the algorithm is used.
     */
    public ReversiAI() {
        ReversiAIbuilder((byte)8, (byte)8, (byte)-1);
    }

    /** Creates a new instance of ReversiAI with
     *  a heightxwidth board and p depth for the
     *  algorithm.
     */
    public void ReversiAIbuilder(byte height, byte width, byte p) {
        plys = p;
        turn = GameField.WHITE;
        
        GameBoard data = new GameBoard(width,height);

        data.setFirstPlayer((byte)3,(byte)3);
        data.setFirstPlayer((byte)4,(byte)4);
        data.setSecondPlayer((byte)3,(byte)4);
        data.setSecondPlayer((byte)4,(byte)3);

        this.tree = new ReversiTree(null, data, null, width, height, turn, new Point(-1,-1));
    }

    /** Returns a table with the current game status
     */
    public GameBoard getCurrentStatus() {
        return this.tree.getBoard();
    }

    /** Returns the current top-level node of the game tree
     */
    public ReversiTree getTree() {
        return this.tree;
    }

    /** Returns board height
     */
    public int getHeight() {
        return this.tree.getBoard().getHeight();
    }

    /** Returns board width
     */
    public int getWidth() {
        return this.tree.getBoard().getWidth();
    }

    /** Returns the "plys" used in minimax algorithm by the AI
     */
    public int getPlys() {
        return this.plys;
    }

    public void setHumanMove(Point p){
        this.humanMove = p;
    }
    
    public Point getHumanMove(){
        return this.humanMove;
    }
    
    /** Returns the current turn
     */
    public byte getTurn() {
        return this.turn;
    }

    /** Changes player's turn to the next one
     */
    public void changeTurn() {
        if (this.turn == GameField.WHITE)
            this.turn = GameField.BLACK;
        else
            this.turn = GameField.WHITE;
    }

    /** Set the "plys" to use by the AI
     */
    public void setPlys(byte p) {
        this.plys  = p;
        this.uplys = p;
    }
    
    public boolean isSinglePlayer(){
        return this.singlePlayer;
    }
    
    public void setSinglePlayer(boolean p) {
        this.singlePlayer = p;
    }

    public void setGui(ReversiGUI g) {
        this.gui = g;
    }

    /** Determines wheter or not a move is valid within
     *  reversi game rules
     */
    public boolean isValidMove(byte row, byte col) {
        Set s = new LinkedHashSet();
        return tree.isValidTableMove(row,col,this.turn,s);
    }
    
    /** returns whether or not a player could make at least one legal
     *  move
     */
    public boolean canMakeMove() {
        ArrayList sons = this.tree.getSons();

        return (this.turn == this.tree.getTurn());
    }

    /** Update the game tree to reflect the current status
     *  when a move was done in (x,y).
     */
    public void update(byte x, byte y) {
        ArrayList a = this.tree.getSons();
        ReversiTree n;
        Set s = new LinkedHashSet();
        boolean found = false;
        byte turn = this.tree.getTurn(), next;
        GameBoard b = (GameBoard)this.tree.getBoard().clone();

        if((a.isEmpty())||(!tree.getBoard().isEmpty(x,y)))
            return;

        /* Search the sons and take the one with the move
         * specified. 
         */
        for(int i = 0; i < a.size(); i++){
            n = (ReversiTree)a.get(i);
            if (turn == n.getBoard().get(x,y)){
                tree = n;
                found = true;
                break;
            }
        }
        if (!found) {
            if (tree.isValidTableMove(x, y, turn, s)){
                if (turn == GameField.WHITE) {
                    next = GameField.BLACK;
                    b.setFirstPlayer(x,y);
                } else {
                    next = GameField.WHITE;
                    b.setSecondPlayer(x,y);
                }
                Iterator it = s.iterator();
                while (it.hasNext()) {
                    Point p = (Point)it.next();
                    x = (byte)p.getX();
                    y = (byte)p.getY();
                    if (b.isFirstPlayer(x,y))
                        b.setSecondPlayer(x,y);
                    else
                        b.setFirstPlayer(x,y);
                }
                this.tree = new ReversiTree(this.tree,b,s,
                                            b.getWidth(),b.getHeight(),
                                            next,new Point(x,y));
                this.tree.expandir(next);
            }
        }

        System.gc();    // invoke garbage collector
        return;
    }
 
    /** Suggest the best possible move given the current game status
     */
    public Point suggest() {
        int sc, g = 0;
        ArrayList l;
        Iterator i;
        ReversiTree n;

        if (uplys > plys)
            plys++; // if the ply has been reduced because of memory
                    // usage, try to restore it slowly
        sc = minimax(this.tree, 0, (short)-(Short.MAX_VALUE),Short.MAX_VALUE);
        l = this.tree.getSons();
        if (l.isEmpty()) {
            return new Point(-1,-1);
        }
            
        i = l.iterator();
        while (i.hasNext()) {
            n = (ReversiTree)(i.next());

            if (n.getScore() == sc)
                return n.getMove();
            g++;
        }
        return new Point(-1,-1);
    }
    
    
    private short max(short a, short b){
        if (a > b)
            return a;
        return b;
    }
    
    private short min(short a, short b){
        if (a < b)
            return a;
        return b;
    }
    private short minimax(ReversiTree tree, int depth,short alpha, short beta) {
        
        Iterator i;
        ReversiTree n;
        short s = 0, ex = 0;

        try {
            if (plys > depth) {
                // tree has to grow
                if (tree.expandir(tree.getTurn()) != GameField.EMPTY) {
                    i = tree.getSons().iterator();
                    while (i.hasNext()) {
                        // we assume the AI plays with BLACK
                        if (tree.getTurn() == GameField.BLACK){ // MAX
                            n = (ReversiTree)(i.next());
                            s = minimax(n, depth + 1,alpha, beta);
                            if ((s > ex)||(ex == 0))
                                ex = s;
                            alpha = max(alpha, ex);
                            if (alpha >= beta){
                                ArrayList sons = new ArrayList();
                                sons.add(n);
                                tree.setSons(sons);
                                tree.setScore(ex);
                                return ex;
                            }
                        } else { // MIN
                            n = (ReversiTree)(i.next());
                            s = minimax(n, depth + 1,alpha, beta);
                            if ((s <= ex)||(ex == 0))
                                ex = s;
                            beta = min(ex,beta);
                            if (beta <= alpha){
                                ArrayList sons = new ArrayList();
                                sons.add(n);
                                tree.setScore(ex);
                                tree.setSons(sons);
                                return ex;
                            }
                        }
                    }
                } else {
                    ex = e(tree);
                    if (!tree.isFinal())
                        tree.getFather().getSons().remove(this);
                }
            } else {
                ex = e(tree);
                if (!tree.isFinal())
                    tree.getFather().getSons().remove(this);
            }
        } catch (java.lang.OutOfMemoryError e) {
            this.plys--;
            ArrayList l = new ArrayList();
            tree.setSons(l);
            System.gc();
            if (ex == 0)
                ex = e(tree);
        }
        tree.setScore(ex);

        if (gui.getProgress() < 100/(depth+1))
            gui.setProgress(100/(depth+1));

        return ex;
    }

    /* Evaluation function for a board and a player.
     * Evaluation consists in adding player's token weights
     * and subtracting opponent one's.
     */
    private short e(ReversiTree node){
        byte mine, other, mine_border, other_border, mine_corner, other_corner;
        int m_weight, o_weight;

        m_weight = o_weight = mine = other = mine_border = other_border = mine_corner = other_corner = 0;
        for (byte i = 0; i < getHeight(); i++) {
            for(byte j = 0; j < getWidth(); j++) {
                // borders
                if (node.getBoard().get(i,j) == GameField.BLACK &&
                    (i == 1 || i == getHeight() - 1 || j == 1 || j == getWidth() - 1))
                    mine_border++;
                if (node.getBoard().get(i,j) == GameField.WHITE &&
                    (i == 1 || i == getHeight() - 1 || j == 1 || j == getWidth() - 1))
                    other_border++;
                // corners
                if (node.getBoard().get(i,j) == GameField.BLACK &&
                    ((i == 1 && j == 1) || (i == 1 && j == getWidth() - 1) ||
                     (i == getHeight() - 1 && j == 1) ||
                     (i == getHeight() - 1 && j == getWidth() - 1)))
                    mine_corner++;
                if (node.getBoard().get(i,j) == GameField.WHITE &&
                    ((i == 1 && j == 1) || (i == 1 && j == getWidth() - 1) ||
                     (i == getHeight() - 1 && j == 1) ||
                     (i == getHeight() - 1 && j == getWidth() - 1)))
                    other_corner++;
                // totals
                if (node.getBoard().get(i,j) == GameField.BLACK) {
                    m_weight += weights[i][j];
                    mine++;
                }
                if (node.getBoard().get(i,j) == GameField.WHITE) {
                    o_weight += weights[i][j];
                    other++;
                }
            }
        }
        if (node.isFinal() && mine > other) {
            return Short.MAX_VALUE;
        }
        if (node.isFinal() && (mine < other)) {
            return Short.MIN_VALUE;
        }

        return (short)(m_weight + 1.5*m_weight*mine_border + 3*m_weight*mine_corner -
               -o_weight - 1.5*o_weight*other_border - 3*o_weight*other_corner);
    }
}
